import torch
import torch.nn as nn

from .ConditionTransformer import ConditionTransformer
from .SDE import VESDE, VPSDE
from .Sampler import Sampler
from .Trainer import Trainer
from .MultiObsSampler import MultiObsSampler

#################################################################################################
# ///////////////////////////////////// Diffusion Model /////////////////////////////////////////
#################################################################################################

class ScoreBasedInferenceModel(nn.Module):
    def __init__(
            self,
            nodes_size,         # Number of features in joint (theta, x)
            sde_type="vesde",   # Type of Stochastic Differential Equation (SDE)
            sigma=25.0,         # Variance for VESDE
            hidden_size=128,    # Hidden size of the transformer
            depth=6,            # Number of transformer blocks
            num_heads=16,       # Number of attention heads
            mlp_ratio=4,        # Ratio of MLP hidden size to embedding size
            device="cpu"        # Device to initialize the model on
            ):
        
        super(ScoreBasedInferenceModel, self).__init__()

        self.nodes_size = nodes_size
        self.sde_type = sde_type
        self.sigma = sigma
        self.hidden_size = hidden_size
        self.depth = depth
        self.num_heads = num_heads
        self.mlp_ratio = mlp_ratio

        # initialize SDE
        self.sigma = sigma
        if sde_type == "vesde":
            self.sde = VESDE(sigma=self.sigma)
        elif sde_type == "vpsde":
            self.sde = VPSDE()
        else:
            raise ValueError("Invalid SDE type")
        
        # Define Model
        self.model = ConditionTransformer(nodes_size=self.nodes_size, hidden_size=hidden_size, 
                                       depth=depth, num_heads=num_heads, mlp_ratio=mlp_ratio)
        self.model.to(device)
        
        # Define Trainer
        self.trainer = Trainer(self)

        # Define Sampler
        self.sampler = Sampler(self)
        self.multi_obs_sampler = MultiObsSampler(self)
        
    #############################################
    # ----- Forward Diffusion -----
    #############################################

    def forward_diffusion_sample(self, x_0, t, x_1=None, condition_mask=None):
        # Diffusion process for time t with defined SDE
        
        if condition_mask is None:
            condition_mask = torch.zeros_like(x_0)

        if x_1 is None:
            x_1 = torch.randn_like(x_0)*(1-condition_mask)+(condition_mask)*x_0

        std = self.sde.marginal_prob_std(t).reshape(-1, 1).to(x_0.device)
        x_t = x_0 + std * x_1 * (1-condition_mask)
        return x_t
    
    #############################################
    # ----- Output Scaling -----
    #############################################

    def output_scale_function(self, t, x):
        scale = self.sde.marginal_prob_std(t).to(x.device)
        return x / scale
    
    #############################################
    # ----- Training -----
    #############################################
    
    def train(self, theta, x, theta_val=None, x_val=None, 
                batch_size=128, max_epochs=500, lr=1e-3, device="cpu", 
                verbose=True, path=None, name="Model", early_stopping_patience=20):
        """
        Train the model on the provided data

        Args:
            theta: Training Parameters 
            x: Training observations
            batch_size: Batch size for training
            max_epochs: Maximum number of training epochs
            lr: Learning rate
            device: Device to run training on
                    if "cuda", training will be distributed across all available GPUs
            theta_val: Validation Parameters
            x_val: Validation observations
            verbose: Whether to show training progress
            path: Path to save model
            early_stopping_patience: Number of epochs to wait before early stopping
        """

        # Combine theta and x into a single tensor
        # train_data: (num_samples, node_size)
        train_data = torch.cat([theta, x], dim=1)
        if theta_val is not None and x_val is not None:
            val_data = torch.cat([theta_val, x_val], dim=1)

        if device == "cuda":
            world_size = torch.cuda.device_count()
        else :
            world_size = 1

        self.trainer.train(world_size=world_size, train_data=train_data, val_data=val_data,
                            max_epochs=max_epochs, early_stopping_patience=early_stopping_patience, batch_size=batch_size, lr=lr,
                            path=path, name=name, device=device, verbose=verbose)

    #############################################
    # ----- Sample -----
    #############################################
        
    def sample(self, theta=None, x=None, err=None, condition_mask=None,
               timesteps=50, eps=1e-3, num_samples=1000, cfg_alpha=None, multi_obs_inference=False, hierarchy=None,
               order=2, snr=0.1, corrector_steps_interval=5, corrector_steps=5, final_corrector_steps=3,
               device="cpu", verbose=True, method="dpm", save_trajectory=False):
        """
        Sample from the model using the specified method

        Args:
            data: Input data : Tensor of shape (num_samples, num_observed_features)
            condition_mask: Binary mask indicating observed values (1) and latent values (0)
                    Shape: (num_samples, num_total_features)
            timesteps: Number of diffusion steps
            eps: End time for diffusion process
            num_samples: Number of samples to generate
            cfg_alpha: Classifier-free guidance strength

            - DPM-Solver parameters -
            order: Order of DPM-Solver (1, 2 or 3)
            snr: Signal-to-noise ratio for Langevin steps
            corrector_steps_interval: Interval for applying corrector steps
            corrector_steps: Number of Langevin MCMC steps per iteration
            final_corrector_steps: Extra correction steps at the end

            - Other parameters -
            device: Device to run sampling on
            verbose: Whether to show progress bar
            method: Sampling method to use (euler, dpm)
            save_trajectory: Whether to save the intermediate denoising trajectory
        """

        # Combine data and create condition mask
        # 0 for latent values, 1 for observed values
        data = None
        if theta is None and x is not None:
            # If only x is provided, use it as the data
            data = x
            if condition_mask is None:
                condition_mask = torch.cat([torch.zeros(self.nodes_size-data.shape[-1]),torch.ones(data.shape[-1])])
        if x is None and theta is not None:
            # If only theta is provided, use it as the data
            data = theta
            if condition_mask is None:
                condition_mask = torch.cat([torch.ones(data.shape[-1]),torch.zeros(self.nodes_size-data.shape[-1])])
        if theta is not None and x is not None:
            # Both theta and x are provided, raise an error
            raise ValueError("Both theta and x are provided!\nPlease provide theta to sample from the Likelihood OR x to sample from the Posterior.")
        if data is None:
            # If neither theta nor x is provided, infer both
            condition_mask = torch.zeros(self.nodes_size)
            data = torch.zeros(1, self.nodes_size)
            

        # Choose device        
        if device == "cuda":
            # Run sampling on all available GPUs
            world_size = torch.cuda.device_count()
        else:
            # Run sampling on specified device
            world_size = 1
            
        if multi_obs_inference == False:
            samples = self.sampler.sample(world_size=world_size, data=data, err=err, condition_mask=condition_mask, timesteps=timesteps, num_samples=num_samples, device=device, cfg_alpha=cfg_alpha,
                                    order=order, snr=snr, corrector_steps_interval=corrector_steps_interval, corrector_steps=corrector_steps, final_corrector_steps=final_corrector_steps,
                                    verbose=verbose, method=method, save_trajectory=save_trajectory)
            
        elif multi_obs_inference == True:
            # Hierarchical Compositional Score Modeling
            samples = self.multi_obs_sampler.sample(world_size=world_size, data=data, condition_mask=condition_mask, timesteps=timesteps, num_samples=num_samples, device=device, cfg_alpha=cfg_alpha, hierarchy=hierarchy,
                                      order=order, snr=snr, corrector_steps_interval=corrector_steps_interval, corrector_steps=corrector_steps, final_corrector_steps=final_corrector_steps,
                                      verbose=verbose, method=method, save_trajectory=save_trajectory)

        # just return the sampled values
        samples = samples[:,:,(1-condition_mask).bool()] 
        
        return samples
    
    #############################################
    # ----- Save & Load -----
    #############################################
    
    def save(self, path, name="Model"):
        state_dict = {
                'model_state_dict' : self.model.state_dict(),
                'nodes_size': self.nodes_size,
                'sde_type': self.sde_type,
                'sigma': self.sigma,
                'hidden_size': self.hidden_size,
                'depth': self.depth,
                'num_heads': self.num_heads,
                'mlp_ratio': self.mlp_ratio
            }
        
        torch.save(state_dict, f"{path}/{name}.pt")
        
    @staticmethod
    def load(path, device=None):

        if device is None:
            # Automatically detect device
            device = 'cuda' if torch.cuda.is_available() else 'cpu'

        checkpoint = torch.load(path, map_location=device)

        model = ScoreBasedInferenceModel(
            nodes_size=checkpoint['nodes_size'],
            sde_type=checkpoint['sde_type'],
            sigma=checkpoint['sigma'],
            hidden_size=checkpoint['hidden_size'],
            depth=checkpoint['depth'],
            num_heads=checkpoint['num_heads'],
            mlp_ratio=checkpoint['mlp_ratio']
        )

        model.model.load_state_dict(checkpoint['model_state_dict'])

        return model
    


import torch
from torch.utils.data import DataLoader, Dataset, DistributedSampler
import torch.distributed as dist
from torch.nn.parallel import DistributedDataParallel as DDP
import torch.multiprocessing as mp

import tqdm
import datetime
import os

class TensorTupleDataset(Dataset):
    def __init__(self, tensor1, tensor2, err=None):

        self.tensor1 = tensor1
        self.tensor2 = tensor2
        self.err = err
        assert len(tensor1) == len(tensor2), "Tensors must have the same length"
        
    def __len__(self):
        return len(self.tensor1)
    
    def __getitem__(self, idx):
        if self.err is not None:
            err = self.err[idx]
        else:
            err = torch.zeros_like(self.tensor1[idx])

        return self.tensor1[idx], self.tensor2[idx], idx, err

#################################################################################################
# ////////////////////////////////////////// Sampling //////////////////////////////////////////
#################################################################################################
class Sampler():
    def __init__(self, SBIm):
        self.SBIm = SBIm
        # Get SDE from model for calculations
        self.sde = self.SBIm.sde

        #############################################
    # ----- Main Sampling Loop -----
    #############################################
    
    def sample(self, world_size, data, err=None, condition_mask=None, 
               timesteps=50, eps=1e-3, num_samples=1000, cfg_alpha=None,
               order=2, snr=0.1, corrector_steps_interval=5, corrector_steps=5, final_corrector_steps=3,
               device="cpu", verbose=True, method="dpm", save_trajectory=False, result_dict=None):
        """
        Sample from the model using the specified method

        Args:
            data: Input data
                    - Should be a DataLoader or a tuple of (data, condition_mask)
                        - Shape data: (num_samples, num_observed_features)
                        - Shape condition_mask: (num_samples, num_total_features)
                    - Can also be a single tensor of data, in which case condition_mask must be provided
            condition_mask: Binary mask indicating observed values (1) and latent values (0)
                    Shape: (num_samples, num_total_features)
            timesteps: Number of diffusion steps
            eps: End time for diffusion process
            num_samples: Number of samples to generate
            cfg_alpha: Classifier-free guidance strength

            - DPM-Solver parameters -
            order: Order of DPM-Solver (1, 2 or 3)
            snr: Signal-to-noise ratio for Langevin steps
            corrector_steps_interval: Interval for applying corrector steps
            corrector_steps: Number of Langevin MCMC steps per iteration
            final_corrector_steps: Extra correction steps at the end

            - Other parameters -
            device: Device to run sampling on
            verbose: Whether to show progress bar
            method: Sampling method to use (euler, dpm)
            save_trajectory: Whether to save the intermediate denoising trajectory
        """

        # Set parameters
        self.world_size = world_size
        self.timesteps = timesteps
        self.eps = eps
        self.num_samples = int(num_samples)
        self.cfg_alpha = cfg_alpha
        self.verbose = verbose
        self.method = method
        self.save_trajectory = save_trajectory

        if method == "dpm":
            self.corrector_steps_interval = corrector_steps_interval
            self.corrector_steps = corrector_steps
            self.final_corrector_steps = final_corrector_steps
            self.snr = snr
            self.order = order

        if self.world_size > 1:
            manager = mp.Manager()
            result_dict = manager.dict()
            mp.spawn(self._sample_loop, args=(data, err, condition_mask, num_samples, result_dict), nprocs=self.world_size, join=True)
            samples = result_dict.get('samples', None)
            manager.shutdown()

        else:
            rank = 0
            self.device = device
            samples = self._sample_loop(rank, data, err, condition_mask, num_samples)

        return samples
   
    def _sample_loop(self, rank, data, err, condition_mask, num_samples, result_dict=None):
        # Set rank
        self.rank = rank
        self.verbose = self.verbose if self.rank == 0 else False

        # Set distributed parameters
        if self.world_size > 1:
            self._ddp_setup(rank, self.world_size)
        else:
            self.model = self.SBIm.model.to(self.device)
        
        # Check data structure
        data_loader, self.num_observations = self._check_data_structure(data, condition_mask, err)

        # Set up timesteps
        self.timesteps_list = torch.linspace(1., self.eps, self.timesteps, device=self.device)
        self.dt = self.timesteps_list[0] - self.timesteps_list[1]
        
        # Loop over data samples
        all_samples = []
        indices = []
        for batch in data_loader:
            # Prepare data for sampling
            data_batch, condition_mask_batch, idx = self._prepare_data(batch, num_samples, self.device)

            # Draw samples from initial noise distribution
            data_batch = self._initial_sample(data_batch, condition_mask_batch)
            
            # Get samples for this batch
            if self.method == "euler":
                samples = self._basic_sampler(data_batch, condition_mask_batch)
            elif self.method == "dpm":
                samples = self._dpm_sampler(data_batch, condition_mask_batch,
                                            order=self.order, snr=self.snr, corrector_steps_interval=self.corrector_steps_interval,
                                            corrector_steps=self.corrector_steps, final_corrector_steps=self.final_corrector_steps)
            elif self.method == "multi_observation":
                samples = self._multi_observation_sampler(data_batch, condition_mask_batch)
            else:
                raise ValueError(f"Sampling method {self.method} not recognized.")

            # Store samples
            all_samples.append(samples)
            indices.append(idx)
            
        # Collect results from all processes if distributed
        if self.world_size > 1:
            dist.barrier()
            self._gather_samples(all_samples, indices, result_dict)
            dist.barrier()
            dist.destroy_process_group()

        else:
            samples = torch.cat(all_samples, dim=0)
            return samples

    #############################################
    # ----- Multi-GPU setup -----
    #############################################

    def _ddp_setup(self, rank, world_size):
        # Setup DistributedDataParallel
        os.environ['MASTER_ADDR'] = 'localhost'
        os.environ["MASTER_PORT"] = "29500"

        torch.cuda.set_device(rank)
        dist.init_process_group(
            backend='nccl',
            init_method='env://',
            world_size=world_size,
            rank=rank,
            timeout=datetime.timedelta(seconds=100_000_000)
        )

        self.SBIm.model.eval()
        self.device = torch.device(f'cuda:{rank}')
        self.SBIm.model.to(self.device)
        self.model = DDP(self.SBIm.model, device_ids=[rank])
            
    def _gather_samples(self, all_samples, indices, result_dict):
        # Gather samples from all processes
        # Convert the list of tensors to a single tensor
        samples = torch.cat(all_samples, dim=0).to(self.device)
        indices = torch.cat(indices, dim=0).to(self.device)

        # Create empty tensors to gather results across processes
        gathered_samples = [torch.zeros_like(samples) for _ in range(self.world_size)]
        gathered_idx = [torch.zeros_like(indices) for _ in range(self.world_size)]

        # Gather data from all processes
        dist.all_gather(gathered_samples, samples)
        dist.all_gather(gathered_idx, indices)

        if self.rank == 0:
            # Sort results by index
            gathered_idx = torch.cat(gathered_idx, dim=0)
            gathered_samples = torch.cat(gathered_samples, dim=0)
            unique_sort_idx = [(gathered_idx == i).nonzero()[0,0].tolist() for i in gathered_idx.unique()]
            samples = gathered_samples[unique_sort_idx]

            result_dict['samples'] = samples.cpu()
        
    #############################################
    # ----- Standard Functions -----
    #############################################

    def _get_score(self, x, t, condition_mask, cfg_alpha=None):
        """Get score estimate with optional classifier-free guidance"""
        # Get conditional score
        with torch.no_grad():
            score_cond = self.model(x=x, t=t, c=condition_mask)
            score_cond = self.SBIm.output_scale_function(t, score_cond)
            
            # Apply classifier-free guidance if requested
            if cfg_alpha is not None:
                score_uncond = self.model(x=x, t=t, c=torch.zeros_like(condition_mask))
                score_uncond = self.SBIm.output_scale_function(t, score_uncond)
                score = score_uncond + cfg_alpha * (score_cond - score_uncond)
            else:
                score = score_cond
                
        return score

    def _check_data_shape(self, data, condition_mask, err):
        # Check data shape
        # Required shape: (num_samples, num_features)
        if len(data.shape) == 1:
            data = data.unsqueeze(0)

        # Check condition mask shape
        # Required shape: (num_samples, num_features)
        if len(condition_mask.shape) == 1:
            condition_mask = condition_mask.unsqueeze(0).repeat(data.shape[0], 1)

        # Check error shape
        # Required shape: (num_samples, num_features)
        if err is not None and len(err.shape) == 1:
            err = err.unsqueeze(0).repeat(data.shape[0], 1)

        return data, condition_mask, err

    def _check_data_structure(self, data, condition_mask, err, batch_size=1e3):
        # Convert data to DataLoader
        
        data, condition_mask, err = self._check_data_shape(data, condition_mask, err)
        dataset_cond = TensorTupleDataset(data, condition_mask, err)
        
        # Use DistributedSampler for multi-GPU
        if self.world_size > 1:
            sampler = DistributedSampler(
                dataset_cond, 
                num_replicas=self.world_size, 
                rank=self.rank,
                shuffle=False,
                drop_last=False,
            )
            data_loader = DataLoader(
                dataset_cond, 
                batch_size=int(batch_size), 
                sampler=sampler, 
                pin_memory=True,
                shuffle=False,
                drop_last=False
            )
        else:
            data_loader = DataLoader(dataset_cond, batch_size=int(batch_size), shuffle=False)
        
        num_observations = dataset_cond.__len__()

        return data_loader, num_observations

    def _prepare_data(self, batch, num_samples, device):
        # Expand data and condition mask to match num_samples
        data, condition_mask, idx, err = batch
        data = data.to(device)
        condition_mask = condition_mask.to(device)

        data = data.unsqueeze(1).repeat(1,num_samples,1)
        condition_mask = condition_mask.unsqueeze(1).repeat(1,num_samples,1)
        err = err.unsqueeze(1).repeat(1,num_samples,1)

        if err is not None:
            # Add scatter around the data
            err = err.to(device)
            data = torch.normal(data, err)

        joint_data = torch.zeros_like(condition_mask)
        if torch.sum(condition_mask==1).item()!=0:
            joint_data[condition_mask==1] = data.flatten()

        return joint_data, condition_mask, idx
    
    def _initial_sample(self, data, condition_mask):
        # Initialize with noise
        # Draw samples from initial noise distribution for latent variables
        # Keep observed variables fixed

        random_noise_samples = self.sde.marginal_prob_std(torch.ones_like(data)) * torch.randn_like(data) * (1-condition_mask)

        data += random_noise_samples     
        return data
  
    #############################################
    # ----- Basic Sampling -----
    #############################################
    
    # Euler-Maruyama sampling
    def _basic_sampler(self, data, condition_mask):
        """
        Basic Euler-Maruyama sampling method
        
        Args:
            data: Input data 
                    Shape: (batch_size, num_samples, num_features)
            condition_mask: Binary mask indicating observed values (1) and latent values (0)
                    Shape: (batch_size, num_samples, num_features)  
        """

        if self.save_trajectory:
            # Storage for trajectory (optional)
            self.data_t = torch.zeros(data.shape[0], self.timesteps+1, data.shape[1], data.shape[2])
            self.score_t = torch.zeros(data.shape[0], self.timesteps+1, data.shape[1], data.shape[2])
            self.dx_t = torch.zeros(data.shape[0], self.timesteps+1, data.shape[1], data.shape[2])
            self.data_t[:,0,:,:] = data
            
        # Main sampling loop
        for n in tqdm.tqdm(range(len(data)), disable=not self.verbose):
            for i, t in enumerate(self.timesteps_list):

                t = t.reshape(-1, 1)
                
                # Get score estimate
                score = self._get_score(data[n,:], t, condition_mask[n,:], self.cfg_alpha)
                
                # Update step
                dx = self.sde.sigma**(2*t) * score * self.dt
                
                # Apply update respecting condition mask
                data[n,:] = data[n,:] + dx * (1-condition_mask[n,:])
                
                if self.save_trajectory:
                    # Store trajectory data
                    self.data_t[n,i+1] = data[n,:]
                    self.dx_t[n,i] = dx
                    self.score_t[n,i] = score

        return data.detach()
    
    #############################################
    # ----- Advanced Sampling -----
    #############################################
    
    # DPM sampling with Langevin corrector steps

    def _corrector_step(self, x, t, condition_mask, steps, snr, cfg_alpha=None):
        """
        Corrector steps using Langevin dynamics
        
        Args:
            x: Input data
            t: Time step
            """
        for _ in range(steps):
            # Get score estimate
            score = self._get_score(x, t, condition_mask, cfg_alpha)
            
            # Langevin dynamics update
            noise_scale = torch.sqrt(snr * 2 * self.sde.marginal_prob_std(t)**2)
            noise = torch.randn_like(x) * noise_scale
            
            # Update x with the score and noise, respecting the condition mask
            grad_step = snr * self.sde.marginal_prob_std(t)**2 * score
            x = x + grad_step * (1-condition_mask) + noise * (1-condition_mask)

        return x

    def _dpm_solver_1_step(self, data_t, t, t_next, condition_mask):
        """First-order solver"""
        sigma_now = self.sde.sigma_t(t)

        # First-order step
        score_now = self._get_score(data_t, t, condition_mask, self.cfg_alpha)
        data_next = data_t + (t-t_next) * sigma_now * score_now * (1-condition_mask)

        return data_next
    
    def _dpm_solver_2_step(self, data_t, t, t_next, condition_mask):
        """Second-order solver"""
        sigma_now = self.sde.sigma_t(t)
        sigma_next = self.sde.sigma_t(t_next)

        # First-order step
        score_half = self._get_score(data_t, t, condition_mask, self.cfg_alpha)
        data_half = data_t + (t-t_next) * sigma_now * score_half * (1-condition_mask)

        # Second-order step
        score_next = self._get_score(data_half, t_next, condition_mask, self.cfg_alpha)
        data_next = data_t + 0.5 * (t-t_next) * (sigma_now**2 * score_half + sigma_next**2 * score_next) * (1-condition_mask)

        return data_next
    
    def _dpm_solver_3_step(self, data_t, t, t_next, condition_mask):
        """Third-order solver"""
        # Get sigma values at different time points
        sigma_t = self.sde.sigma_t(t)
        t_mid = (t + t_next) / 2
        sigma_mid = self.sde.sigma_t(t_mid)
        sigma_next = self.sde.sigma_t(t_next)

        # First calculate the intermediate score at time t
        score_t = self._get_score(data_t, t, condition_mask, self.cfg_alpha)
        
        # First intermediate point (Euler step)
        data_mid1 = data_t + (t - t_mid) * sigma_t * score_t * (1-condition_mask)
        
        # Get score at the first intermediate point
        score_mid1 = self._get_score(data_mid1, t_mid, condition_mask, self.cfg_alpha)
        
        # Second intermediate point (using first intermediate)
        data_mid2 = data_t + (t - t_mid) * ((1/3) * sigma_t * score_t + (2/3) * sigma_mid * score_mid1) * (1-condition_mask)
        
        # Get score at the second intermediate point
        score_mid2 = self._get_score(data_mid2, t_mid, condition_mask, self.cfg_alpha)
        
        # Final step using all information
        data_next = data_t + (t - t_next) * ((1/4) * sigma_t * score_t + 
                                            (3/4) * sigma_next * score_mid2) * (1-condition_mask)
        
        return data_next

    def _dpm_sampler(self, data, condition_mask, 
                     order=2, 
                     snr=0.1, corrector_steps_interval=5, corrector_steps=5, final_corrector_steps=3):
        """
        Hybrid sampling approach combining DPM-Solver with Predictor-Corrector refinement.

        Args:
            data: Input data
            condition_mask: Binary mask indicating observed values (1) and latent values (0)
            order: Order of DPM-Solver (1, 2 or 3)
            snr: Signal-to-noise ratio for Langevin steps
            corrector_steps_interval: Interval for applying corrector steps
            corrector_steps: Number of Langevin MCMC steps per iteration
            final_corrector_steps: Extra correction steps at the end
        """

        if self.save_trajectory:
            # Storage for trajectory (optional)
            self.data_t = torch.zeros(data.shape[0], self.timesteps, data.shape[1], data.shape[2])
            self.data_t[:,0,:,:] = data

        # Main sampling loop
        for n in tqdm.tqdm(range(len(data)), disable=not self.verbose):
            for i in range(self.timesteps-1):
                
                # ------- PREDICTOR: DPM-Solver -------
                t_now = self.timesteps_list[i].reshape(-1, 1)
                t_next = self.timesteps_list[i+1].reshape(-1, 1)

                if order == 1:
                    data[n,:] = self._dpm_solver_1_step(data[n,:], t_now, t_next, condition_mask[n,:])
                elif order == 2:
                    data[n,:] = self._dpm_solver_2_step(data[n,:], t_now, t_next, condition_mask[n,:])
                elif order == 3:
                    data[n,:] = self._dpm_solver_3_step(data[n,:], t_now, t_next, condition_mask[n,:])
                else:
                    raise ValueError(f"Only orders 1, 2 or 3 are supported in the DPM-Solver.")
                
                # ------- CORRECTOR: Langevin MCMC steps -------
                # Only apply corrector steps occasionally to save computation
                if corrector_steps > 0 and (i % corrector_steps_interval == 0 or i >= self.timesteps - final_corrector_steps):
                    steps = corrector_steps
                    if i >= self.timesteps - final_corrector_steps:
                        steps = corrector_steps * 2  # More steps at the end
                        
                    data[n,:] = self._corrector_step(data[n,:], t_next, condition_mask[n,:], 
                                               steps, snr, self.cfg_alpha)

                if self.save_trajectory:
                    # Store trajectory data
                    self.data_t[n,i+1] = data[n,:]

        return data.detach()
    


import torch
import torch.nn as nn
import math

#################################################################################################
# ///////////////////////////////////////// Transformer /////////////////////////////////////////
#################################################################################################

"""
    Transformer model for diffusion models.
    
    - modulate: Modulates the input with shift and scale.
    - InputEmbedder: Embeds joint data into vector representations.
    - TimestepEmbedder: Embeds scalar timesteps into vector representations.
    - Mlp: MLP for Output of Self-Attention
    - TransformerBlock: A ConditionTransformer block with adaptive layer norm zero (adaLN-Zero) conditioning.
    - FinalLayer: The final layer of ConditionTransformer.
    - ConditionTransformer: Transformer model for diffusion models.
"""

def modulate(x, shift, scale):
    return x * (1 + scale) + shift

#############################################
# ----- Timestep & Input Embedding -----
#############################################

class InputEmbedder(nn.Module):
    """
    Embeds joint data into vector representations.
    """
    def __init__(self, nodes_size, hidden_size):
        super().__init__()
        self.embedding_params = nn.Parameter(torch.ones(1, nodes_size, hidden_size))

    def forward(self, x):
        x = x.unsqueeze(-1) * self.embedding_params
        return x

class TimestepEmbedder(nn.Module):
    """
    Embeds scalar timesteps into vector representations.
    """
    def __init__(self, nodes_size, hidden_size, mlp_ratio=4.0, frequency_embedding_size=256):
        super().__init__()
        self.mlp = nn.Sequential(
            nn.Linear(frequency_embedding_size, mlp_ratio*hidden_size, bias=True),
            nn.SiLU(),
            nn.Linear(mlp_ratio*hidden_size, frequency_embedding_size, bias=True),
        )
        self.frequency_embedding_size = frequency_embedding_size

    @staticmethod
    def timestep_embedding(t, dim, max_period=10000):
        """
        Create sinusoidal timestep embeddings.
        :param t: a 1-D Tensor of N indices, one per batch element.
                          These may be fractional.
        :param dim: the dimension of the output.
        :param max_period: controls the minimum frequency of the embeddings.
        :return: an (N, D) Tensor of positional embeddings.
        """

        half = dim // 2

        freqs = torch.exp(
            -math.log(max_period) * torch.arange(start=0, end=half, dtype=torch.float32) / half
        ).to(device=t.device)

        args = t * freqs
        embedding = torch.cat([torch.cos(args), torch.sin(args)], dim=-1)
        if dim % 2:
            embedding = torch.cat([embedding, torch.zeros_like(embedding[:, :1])], dim=-1)
        return embedding

    def forward(self, t):
        t_freq = self.timestep_embedding(t, self.frequency_embedding_size)
        t_emb = self.mlp(t_freq)
        return t_emb
    
#############################################
# ----- MLP -----
#############################################

class Mlp(nn.Module):
    """
    MLP for Output of Self-Attention
    """
    def __init__(
            self,
            in_features,
            hidden_features=None,
            out_features=None,
            act_layer=nn.GELU,
            bias=True,
    ):
        super().__init__()

        out_features = out_features or in_features
        hidden_features = hidden_features or in_features

        self.fc1 = nn.Linear(in_features, hidden_features, bias=bias)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features, bias=bias)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        x = self.fc2(x)

        return x

#############################################
# ----- Core ConditionTransformer Model -----
#############################################

class TransformerBlock(nn.Module):
    """
    A ConditionTransformer block with adaptive layer norm zero (adaLN-Zero) conditioning.
    """
    def __init__(self, hidden_size, num_heads, nodes_size, mlp_ratio=4.0, time_embedding_size=256, **block_kwargs):
        super().__init__()
        self.hidden_size = hidden_size
        self.nodes_size = nodes_size
        self.num_heads = num_heads

        self.norm1 = nn.LayerNorm((nodes_size, hidden_size), elementwise_affine=False, eps=1e-6)

        self.attn = nn.MultiheadAttention(hidden_size, num_heads=num_heads, add_bias_kv=True, batch_first=True, **block_kwargs )  

        self.norm2 = nn.LayerNorm((nodes_size, hidden_size), elementwise_affine=False, eps=1e-6)

        mlp_hidden_dim = int(hidden_size * mlp_ratio)
        approx_gelu = lambda: nn.GELU(approximate="tanh")
        self.mlp = Mlp(in_features=hidden_size, hidden_features=mlp_hidden_dim, act_layer=approx_gelu)

        self.adaLN_modulation = nn.Sequential(
            nn.SiLU(),
            nn.Linear(time_embedding_size, 6 * nodes_size * hidden_size, bias=True)
        )
    
    def forward(self, x, c, t):
        shift_msa, scale_msa, gate_msa, shift_mlp, scale_mlp, gate_mlp = self.adaLN_modulation(t).reshape(-1, self.nodes_size, 6*self.hidden_size).chunk(6, dim=-1)
        
        x_norm = modulate(self.norm1(x), shift_msa, scale_msa)

        q, k, v = x_norm.repeat(1,1,3).chunk(3, dim=-1)

        # Attention mask to prevent latent nodes from attending to other latent nodes
        attn_mask = (1-c).type(torch.bool).unsqueeze(1).repeat(self.num_heads, self.nodes_size, 1)

        x = x + gate_msa * self.attn(q, k, v, need_weights=False, attn_mask=attn_mask)[0]
        x = x + gate_mlp * self.mlp(modulate(self.norm2(x), shift_mlp, scale_mlp))

        return x

#############################################
# ----- Final Layer -----
#############################################

class FinalLayer(nn.Module):
    """
    The final layer of ConditionTransformer.
    """
    def __init__(self, hidden_size, nodes_size, time_embedding_size=256):
        super().__init__()
        self.hidden_size = hidden_size
        self.nodes_size = nodes_size

        self.norm_final = nn.LayerNorm((nodes_size, hidden_size), elementwise_affine=False, eps=1e-6)
        self.embedding_params = nn.Parameter(torch.zeros(nodes_size*hidden_size, nodes_size))

        self.adaLN_modulation = nn.Sequential(
            nn.SiLU(),
            nn.Linear(time_embedding_size, 2 * nodes_size * hidden_size, bias=True)
        )

    def forward(self, x, t):
        shift, scale = self.adaLN_modulation(t).reshape(-1, self.nodes_size, 2*self.hidden_size).chunk(2, dim=-1)
        x = modulate(self.norm_final(x), shift, scale)
        x = nn.SiLU()(x)
        x = x.flatten(1) @ self.embedding_params
        return x.squeeze(-1)

#############################################
# ----- ConditionTransformer Model -----
#############################################

class ConditionTransformer(nn.Module):
    """
    Diffusion model with a Transformer backbone.
    """
    def __init__(
        self,
        nodes_size,
        hidden_size=64,
        time_embedding_size=256,
        depth=6,
        num_heads=16,
        mlp_ratio=4.0,
    ):
        super().__init__()
        self.nodes_size = nodes_size
        self.num_heads = num_heads
        self.time_embedding_size = time_embedding_size

        self.x_embedder = InputEmbedder(nodes_size=nodes_size, hidden_size=hidden_size)
        self.t_embedder = TimestepEmbedder(nodes_size, hidden_size, mlp_ratio=mlp_ratio, frequency_embedding_size=time_embedding_size)

        self.blocks = nn.ModuleList([
            TransformerBlock(hidden_size, num_heads, nodes_size, mlp_ratio=mlp_ratio) for _ in range(depth)
        ])
        self.final_layer = FinalLayer(hidden_size, nodes_size)                   
        self.initialize_weights()

    def initialize_weights(self):
        # Initialize transformer layers:
        def _basic_init(module):
            if isinstance(module, nn.Linear):
                torch.nn.init.xavier_uniform_(module.weight)
                if module.bias is not None:
                    nn.init.constant_(module.bias, 0)
        self.apply(_basic_init)

        # Initialize timestep embedding MLP:
        nn.init.normal_(self.t_embedder.mlp[0].weight, std=0.02)
        nn.init.normal_(self.t_embedder.mlp[2].weight, std=0.02)

        # Zero-out adaLN modulation layers in ConditionTransformer blocks:
        for block in self.blocks:
            nn.init.constant_(block.adaLN_modulation[-1].weight, 0)
            nn.init.constant_(block.adaLN_modulation[-1].bias, 0)

        # Zero-out output layers:
        nn.init.constant_(self.final_layer.adaLN_modulation[-1].weight, 0)
        nn.init.constant_(self.final_layer.adaLN_modulation[-1].bias, 0)
        # nn.init.constant_(self.final_layer.linear.weight, 0)
        # nn.init.constant_(self.final_layer.linear.bias, 0)
    
    def forward(self, x, t, c):
        """
        Forward pass of ConditionTransformer.
        x: (N, C, H, W) tensor of spatial inputs (images or latent representations of images)
        t: (N,) tensor of diffusion timesteps
        c: (N,) tensor of data conditions (latent or conditioned)
        """
        x = self.x_embedder(x)
        t = self.t_embedder(t)
        #c_embed = self.c_embedder(c.type(torch.int))
        #t += c_embed
       
        for block in self.blocks:
            x = block(x, c, t)
        
        x = self.final_layer(x, t)

        return x



import os
import sys
import pickle

import torch
import torch.nn as nn

import numpy as np

import scipy
from scipy import optimize
from scipy.stats import norm, gaussian_kde

import matplotlib.pyplot as plt
import seaborn as sns

from .ScoreBasedInferenceModel import ScoreBasedInferenceModel as SBIm

#################################################################################################
# ///////////////////////////////////// Model Comparison ////////////////////////////////////////
#################################################################################################

class ModelTransfuser():
    def __init__(self, path=None):
        
        ## Check if the path exists
        if path is not None:
            if not os.path.exists(path):
                os.makedirs(path)
        self.path = path
        
        self.models_dict = {}
        self.data_dict = {}
        self.trained_models = False # Flag to check if models are trained

    #############################################
    # ----- Model Management -----
    #############################################

    # Add a trained model to the transfuser
    def add_model(self, model_name, model):
        """
        Add a trained model to the transfuser.

        Args:
            model_name: The name of the model.
            model: The model itself.
        """
        self.models_dict[model_name] = model
        self.trained_models = True
        print(f"Model {model_name} added to transfuser.")

    # Add multiple trained models to the transfuser
    def add_models(self, models_dict):
        """
        Add multiple trained models to the transfuser.

        Args:
            models_dict: A dictionary of models to add.
        """
        for model_name, model in models_dict.items():
            self.add_model(model_name, model)

        self.trained_models = True
        print("All models added to transfuser.")

    # Add data to a model
    def add_data(self, model_name, theta, x, val_theta=None, val_x=None):
        """
        Add training and validation data to a model.

        Args:
            model_name: The name of the model.
            train_data: The training data.
            val_data: The validation data (optional).
        """
        if val_theta is None:
            self.data_dict[model_name] = {
                "train_theta": theta,
                "train_x": x,
            }
        else:
            self.data_dict[model_name] = {
                "train_theta": theta,
                "train_x": x,
                "val_theta": val_theta,
                "val_x": val_x,
            }
        self.trained_models = False
        print(f"Data added to model {model_name}")

    # Remove a model from the transfuser
    def remove_model(self, model_name):
        """
        Remove a model from the transfuser.

        Args:
            model_name: The name of the model to remove.
        """
        if model_name in self.models_dict:
            del self.models_dict[model_name]
            print(f"Model {model_name} removed from transfuser.")
        else:
            print(f"Model {model_name} not found in transfuser.")

    #############################################
    # ----- Initialize Models -----
    #############################################

    def init_models(self, sde_type, sigma, hidden_size, depth, num_heads, mlp_ratio):
        """
        Initialize the Score-Based Inference Models with the given parameters

        Args:
            sde_type: The type of SDE
            sigma: The sigma value
            hidden_size: The size of the hidden layer
            depth: The depth of the model
            num_heads: The number of heads in the model
            mlp_ratio: The MLP ratio
        """

        if self.trained_models:
            print("Models are already trained. This will overwrite the models.")
            return
        else:
            init_models = []
            for model_name in self.data_dict.keys():
                nodes_size = self.data_dict[model_name]["train_theta"].shape[1] + self.data_dict[model_name]["train_x"].shape[1]
                self.models_dict[model_name] = SBIm(nodes_size=nodes_size,
                                                    sde_type=sde_type,
                                                    sigma=sigma,
                                                    hidden_size=hidden_size,
                                                    depth=depth,
                                                    num_heads=num_heads,
                                                    mlp_ratio=mlp_ratio)
                init_models.append(model_name)

            print(f"Models initialized: {init_models}")

    #############################################
    # ----- Train Models -----
    #############################################

    def train_models(self, batch_size=128, max_epochs=500, lr=1e-3, device="cuda",
                verbose=False, path=None, early_stopping_patience=20): 
        
        """
        Train the models on the provided data

        Args:
            batch_size: Batch size for training
            max_epochs: Maximum number of training epochs
            lr: Learning rate
            device: Device to run training on
                    if "cuda", training will be distributed across all available GPUs
            verbose: Whether to show training progress
            path: Path to save model
            early_stopping_patience: Number of epochs to wait before early stopping
        """

        if self.trained_models:
            print("Continue training existing models.")

        if path is not None:
            self.path = path
        elif self.path is not None:
            path = self.path

        for model_name in self.models_dict.keys():
            model = self.models_dict[model_name]

            theta = self.data_dict[model_name]["train_theta"]
            x = self.data_dict[model_name]["train_x"]
            val_theta = self.data_dict[model_name].get("val_theta", None)
            val_x = self.data_dict[model_name].get("val_x", None)

            model.train(theta=theta, x=x, theta_val=val_theta, x_val=val_x,
                        batch_size=batch_size, max_epochs=max_epochs, lr=lr, device=device,
                        verbose=verbose, path=path, name=model_name ,early_stopping_patience=early_stopping_patience)

            load_path = f"{path}/{model_name}.pt"
            self.models_dict[model_name] = SBIm.load(path=load_path, device="cpu")
            print(f"Model {model_name} trained")
            torch.cuda.empty_cache()

        self.trained_models = True

    #############################################
    # ----- Model Comparison -----
    #############################################

    def compare(self, x, err=None, condition_mask=None,
               timesteps=50, eps=1e-3, num_samples=1000, cfg_alpha=None, multi_obs_inference=False, hierarchy=None,
               order=2, snr=0.1, corrector_steps_interval=5, corrector_steps=5, final_corrector_steps=3,
               device="cuda", verbose=False, method="dpm"):
        """
        Compare the models on the provided observations

        Args:
            observations: The observations to compare the models on
            condition_mask: Binary mask indicating observed values (1) and latent values (0)
                    Shape: (num_samples, num_total_features)
                    Optional
            timesteps: Number of timesteps for the model
            eps: Epsilon value for the model
            num_samples: Number of samples to generate
            cfg_alpha: CFG alpha value for the model
            multi_obs_inference: Whether to use multi-observation inference
            hierarchy: Hierarchy for the model
            order: Order of the model
            snr: Signal-to-noise ratio for the model
            corrector_steps_interval: Corrector steps interval for the model
            corrector_steps: Corrector steps for the model
            final_corrector_steps: Final corrector steps for the model
            device: Device to run inference on
            verbose: Whether to show inference progress
            method: Method to use for inference
        """

        if not self.trained_models:
            print("Models are not trained or provided. Please train the models before comparing.")
            return
        
        self.stats = {}
        self.model_null_log_probs = {}
        self.softmax = nn.Softmax(dim=0)
        
        for model_name, model in self.models_dict.items():
            self.stats[model_name] = {}
            if condition_mask is None:
                condition_mask = torch.cat([torch.zeros(model.nodes_size-x.shape[-1]),torch.ones(x.shape[-1])])
            ####################
            # Posterior sampling
            posterior_samples = model.sample(x=x, err=err, condition_mask=condition_mask,
                                            timesteps=timesteps, eps=eps, num_samples=num_samples, cfg_alpha=cfg_alpha,
                                            multi_obs_inference=multi_obs_inference, hierarchy=hierarchy,
                                            order=order, snr=snr, corrector_steps_interval=corrector_steps_interval, corrector_steps=corrector_steps, final_corrector_steps=final_corrector_steps,
                                            device=device, verbose=verbose, method=method)
            posterior_samples = posterior_samples.cpu().numpy()

            # MAP estimation
            theta_hat = np.array([self._map_kde(posterior_samples[i]) for i in range(len(posterior_samples))])
            MAP_posterior, std_MAP_posterior = torch.tensor(theta_hat[:,0], dtype=torch.float), torch.tensor(theta_hat[:,1], dtype=torch.float)

            # Storing MAP and std MAP
            self.stats[model_name]["MAP"] = theta_hat

            ####################
            # Null Hypothesis
            null_samples = model.sample(timesteps=timesteps, eps=eps, num_samples=num_samples, cfg_alpha=cfg_alpha,
                                            multi_obs_inference=multi_obs_inference, hierarchy=hierarchy,
                                            order=order, snr=snr, corrector_steps_interval=corrector_steps_interval, corrector_steps=corrector_steps, final_corrector_steps=final_corrector_steps,
                                            device=device, verbose=verbose, method=method)
            null_samples = null_samples[0,:,condition_mask.bool()].cpu().numpy()

            # Log probability of null hypothesis
            null_log_probs = torch.tensor([self._log_prob(null_samples, obs) for obs in x])
            self.stats[model_name]["log_probs_nullHyp"] = null_log_probs

            ####################
            # Likelihood sampling
            likelihood_samples = model.sample(theta=MAP_posterior, err=std_MAP_posterior, condition_mask=condition_mask,
                                            timesteps=timesteps, eps=eps, num_samples=num_samples, cfg_alpha=cfg_alpha,
                                            multi_obs_inference=multi_obs_inference, hierarchy=hierarchy,
                                            order=order, snr=snr, corrector_steps_interval=corrector_steps_interval, corrector_steps=corrector_steps, final_corrector_steps=final_corrector_steps,
                                            device=device, verbose=verbose, method=method)
            likelihood_samples = likelihood_samples.cpu().numpy()

            # Log probability of likelihood
            log_probs = torch.tensor([self._log_prob(likelihood_samples[i], x[i]) for i in range(len(x))])
            self.stats[model_name]["log_probs"] = log_probs
            self.stats[model_name]["AIC"] = log_probs.sum() 

            # Null Hypothesis test
            self.stats[model_name]["Bayes_Factor_Null_Hyp"] = log_probs.sum() - null_log_probs.sum()

        # Calculate Model Probabilitys from AICs
        aics = [self.stats[model_name]["AIC"] for model_name in self.stats.keys()]
        aics = torch.tensor(aics)
        model_probs = self.softmax(aics)

        # Calculate Probability of each observation
        log_probs = torch.stack([self.stats[model_name]["log_probs"] for model_name in self.stats.keys()])
        probs = self.softmax(log_probs)

        for i, model_name in enumerate(self.stats.keys()):
            self.stats[model_name]["model_prob"] = model_probs[i].item()
            self.stats[model_name]["obs_probs"] = probs[i]

        model_names = list(self.stats.keys())
        best_model = model_names[model_probs.argmax()]
        best_model_prob = 100*model_probs.max()

        # Null Hypothesis test
        best_bayes_factor = self.stats[best_model]["Bayes_Factor_Null_Hyp"]
        hypothesis_test = "could" if best_bayes_factor > 0 else "could not"
        hypothesis_test_strength = self._bayes_factor_strength(best_bayes_factor)

        model_print_length = len(max(model_names, key=len))
        print(f"Probabilities of the models after {len(x)} observations:")
        for model in model_names:
            print(f"{model.ljust(model_print_length)}: {100*self.stats[model]['model_prob']:6.2f} %")
        print()
        print(f"Model {best_model} fits the data best " + 
                f"with a relative support of {best_model_prob:.1f}% among the considered models "+
                f"and {hypothesis_test} reject the null hypothesis{hypothesis_test_strength}.")
        
        if self.path is not None:
            with open(f"{self.path}/model_comp.pkl", "wb") as f:
                pickle.dump(self.stats, f)

    #############################################
    # ----- Kernel Density Estimation -----
    #############################################

    def _log_prob(self, samples, observation):
        """Compute the log probability of the samples"""
        kde = gaussian_kde(samples.T)
        log_prob = kde.logpdf(observation).item()
        return log_prob

    def _map_kde(self, samples):
        """Find the joint mode of the multivariate distribution"""
        kde = gaussian_kde(samples.T)  # KDE expects (n_dims, n_samples)
        
        # Start optimization from the mean
        initial_guess = np.mean(samples, axis=0)
        
        # Use full minimize with multiple dimensions
        result = optimize.minimize(lambda x: -kde(x.reshape(-1, 1)), initial_guess)
        std_devs = np.sqrt(np.diag(kde.covariance))

        return result.x, std_devs
    
    ##############################################
    # ----- Bayes Factor Strenght -----
    ##############################################

    def _bayes_factor_strength(self, BF):
        """
        Calculate the strength of the Bayes factor.

        Args:
            BF: The Log Bayes factor.

        Returns:
            The strength of the Bayes factor as a string.
        """
        hypothesis_test_strength = torch.exp(BF)

        if 1 < hypothesis_test_strength <= 3.2:
            hypothesis_test_strength = " barley"
        elif 3.2 < hypothesis_test_strength <= 10:
            hypothesis_test_strength = " substantially"
        elif 10 < hypothesis_test_strength <= 100:
            hypothesis_test_strength = " strongly"
        elif 100 < hypothesis_test_strength:
            hypothesis_test_strength = " decisively"
        else:
            hypothesis_test_strength = ""

        return hypothesis_test_strength
    
    ##############################################
    # ----- Plotting -----
    ##############################################

    def plots(self, stats_dict=None, path=None, show=True):

        # Check path
        if path is None:
            path = self.path
            if not os.path.exists(path):
                os.makedirs(path)

        # Check stats_dict
        if stats_dict is None:
            stats_dict = self.stats

        model_names = list(self.stats.keys())
        model_probs = torch.tensor([self.stats[model]["model_prob"] for model in model_names])
        model_log_probs = torch.stack([self.stats[model]["log_probs"] for model in model_names])
        model_obs_probs = torch.stack([self.stats[model]["obs_probs"] for model in model_names])

        sns.set_context("paper")

        # Plot violin plot of model probabilities
        plt.figure(figsize=(10, 5))
        sns.violinplot(data=model_obs_probs.T)
        plt.xticks(ticks=range(len(model_names)), labels=model_names)
        plt.title("Model Probabilities")
        plt.xlabel("Model")
        plt.ylabel("Probability of Observations")
        if path is not None:
            plt.savefig(f"{path}/model_probs_violin.png")
        if show:
            plt.show()
        plt.close()

        # Plot updated model probabilities
        plt.figure(figsize=(10, 5))
        plt.plot(torch.arange(1, model_log_probs.shape[1]+1).repeat(len(model_names),1).T,
                    torch.nn.functional.softmax(model_log_probs.cumsum(1),0).T,
                    label=model_names, marker='o', markersize=3, linewidth=1)
        plt.legend()
        plt.xscale("log")
        plt.title("Updated Model Probabilities")
        plt.xlabel("Number of observations")
        plt.ylabel("Model Probability")
        plt.grid(True)
        if path is not None:
            plt.savefig(f"{path}/model_probs_updated.png")
        if show:
            plt.show()
        plt.close()



from .ConditionTransformer import *
from .ModelTransfuser import *
from .ScoreBasedInferenceModel import *
from .SDE import *
from .Trainer import *
from .Sampler import *
from .MultiObsSampler import *


import os
import torch
from torch.utils.data import DataLoader, Dataset, DistributedSampler
import torch.distributed as dist 
from torch.nn.parallel import DistributedDataParallel as DDP
import torch.multiprocessing as mp

import schedulefree

import tqdm
import datetime
import time

class TensorTupleDataset(Dataset):
    def __init__(self, tensor1, tensor2):
        self.tensor1 = tensor1
        self.tensor2 = tensor2
        
    def __len__(self):
        return len(self.tensor1)
    
    def __getitem__(self, idx):
        data = self.tensor1[idx]

        if isinstance(self.tensor2, torch.distributions.Distribution):
            cond_mask = self.tensor2.sample(data.shape)
        else:
            cond_mask = self.tensor2[idx]

        return data, cond_mask, idx
    
#################################################################################################
# ////////////////////////////////////////// Training //////////////////////////////////////////
#################################################################################################
class Trainer():
    def __init__(self, SBIm):
        self.SBIm = SBIm
        #self.model_copy = copy.deepcopy(self.SBIm.model)
        # Get SDE from model for calculations
        self.sde = self.SBIm.sde

    #############################################
    # ----- Training loop -----
    #############################################
    def train(self, world_size, train_data, val_data=None,
              max_epochs=500, early_stopping_patience=20, batch_size=128, lr=1e-3,
              path=None, name="Model", device="cpu", verbose=True):
        
        """
        Training function for the score prediction task

        Args:
            rank: Rank of the current process
            world_size: Number of processes
            train_data: Training data
            val_data: Validation data
            max_epochs: Maximum number of epochs
            early_stopping_patience: Number of epochs to wait before early stopping
            batch_size: Batch size
            lr: Learning rate
            path: Path to save the model
            name: Name of the model
            device: Device to use
            verbose: Verbosity
        """
        start_time = time.time()

        # Create Checkpoint directory
        if path is None:
            path = "data/models/Model_test/"
        if not os.path.exists(path):
            os.makedirs(path)
        
        # Set Parameters
        self.world_size = world_size
        self.max_epochs = max_epochs
        self.early_stopping_patience = early_stopping_patience
        self.batch_size = batch_size
        self.lr = lr
        self.path = path
        self.name = name
        self.name_checkpoint = f"{self.name}_checkpoint"
        self.verbose = verbose
        self.eps = 1e-3 # Epsilon for numerical stability and endpoint in diffusion process

        if self.world_size > 1:
            mp.spawn(self._train_loop, args=(train_data, val_data), nprocs=self.world_size)
        else:
            rank = 0
            self.device = device
            self._train_loop(rank, train_data, val_data)

        end_time = time.time()
        training_time = (end_time-start_time) / 60
        if self.verbose:
            print(f"Training took {training_time:.1f} minutes")

    def _train_loop(self, rank, train_data, val_data):

        # Set device distribution
        if self.world_size > 1:
            self._ddp_setup(rank, self.world_size)
        else:
            self.model = self.SBIm.model.to(self.device)

        self.verbose = self.verbose if rank == 0 else False

        # Check data structure
        data_loader = self._prepare_data(train_data, batch_size=self.batch_size, rank=rank)
        if val_data is not None:
            val_loader = self._prepare_data(val_data, batch_size=1_000, rank=rank)

        # Init tracking variables
        best_val_loss = float('inf')
        patience_counter = 0
        self.train_loss = []
        self.val_loss = []

        # Set optimizer
        optimizer = schedulefree.AdamWScheduleFree(self.model.parameters(), lr=self.lr)

        for epoch in range(self.max_epochs):
            # Train
            if self.world_size > 1: dist.barrier()

            train_loss_process = self._run_epoch(epoch, data_loader, optimizer, is_train=True)
            
            if self.world_size > 1:
                # Wait for all processes to finish training
                dist.barrier()

                # Gather losses from all processes
                epoch_train_loss = [torch.zeros(1).to(self.device) for _ in range(self.world_size)]
                dist.all_gather(epoch_train_loss, torch.tensor(train_loss_process).to(self.device))
                train_loss_all = torch.mean(torch.stack(epoch_train_loss)).item()
            else:
                train_loss_all = train_loss_process

            self.train_loss.append(train_loss_all)

            # Early stopping on training loss if no validation data is provided
            if val_data is None and train_loss_all < best_val_loss:
                best_val_loss = train_loss_all
                patience_counter = 0
                if rank == 0:
                    self._save_checkpoint(name=self.name_checkpoint)
            elif val_data is None and train_loss_all >= best_val_loss:
                patience_counter += 1


            # Validate
            if val_data is not None:
                val_loss_process = self._run_epoch(epoch, val_loader, optimizer, is_train=False)

                if self.world_size > 1:
                    # Wait for all processes to finish validation
                    dist.barrier()

                    # Gather losses from all processes
                    epoch_val_loss = [torch.zeros(1).to(self.device) for _ in range(self.world_size)]
                    dist.all_gather(epoch_val_loss, torch.tensor(val_loss_process).to(self.device))
                    val_loss_all = torch.mean(torch.stack(epoch_val_loss)).item()
                else:
                    val_loss_all = val_loss_process

                self.val_loss.append(val_loss_all)

                # Early stopping
                if val_loss_all < best_val_loss:
                    best_val_loss = val_loss_all
                    patience_counter = 0
                    if rank == 0:
                        self._save_checkpoint(name=self.name_checkpoint)
                elif val_loss_all >= best_val_loss:
                    patience_counter += 1

            # Wait for all processes to finish validation
            if self.world_size > 1: dist.barrier()

            if self.verbose:
                if val_data is not None:
                    print(f'--- Epoch: {epoch+1:3d} --- Training Loss: {train_loss_all:8.3f} --- Validation Loss: {val_loss_all:8.3f} ---')
                else:
                    print(f'--- Epoch: {epoch+1:3d} --- Training Loss: {train_loss_all:8.3f} ---')
                print()
                time.sleep(0.2)

            if patience_counter == self.early_stopping_patience:
                    break
            
            
        if self.world_size > 1:
            dist.barrier()
            if rank == 0:
                self._save_checkpoint(name=self.name)
            dist.destroy_process_group()
        
    def _run_epoch(self, epoch, data_loader, optimizer, is_train):
        if self.world_size > 1:
            data_loader.sampler.set_epoch(epoch)

        if is_train:
            self.model.train()
            optimizer.train()
        else:
            self.model.eval()
            optimizer.eval()

        total_loss = 0
        batch_count = 0

        show_progress = self.verbose if is_train else False
        for batch in tqdm.tqdm(data_loader, disable=not show_progress):
            if is_train:
                optimizer.zero_grad()
            
            loss = self._run_batch(batch)
            total_loss += loss.item() / batch[0].shape[0]
            batch_count += 1

            if is_train:
                loss.backward()
                if self.world_size > 1: dist.barrier()
                optimizer.step()

        return total_loss / batch_count

    def _run_batch(self, batch):
        # Get data
        data, condition_mask, idx = self._prepare_batch(batch, self.device)

        # Get timesteps
        timesteps = torch.rand(data.shape[0], 1, device=self.device) * (1.-self.eps)+self.eps

        # Sample x_1 from noise distribution
        x_1 = torch.randn_like(data)*(1-condition_mask) + data*condition_mask
        # Calculate x at time t in diffusion process
        x_t = self.SBIm.forward_diffusion_sample(data, timesteps, x_1, condition_mask)
        # Get score
        score = self._get_score(x_t, timesteps, condition_mask)
        # Calculate loss
        loss = self.loss_fn(score, timesteps, x_1, condition_mask)

        return loss

    #############################################
    # ----- Loss Function -----
    #############################################
    def loss_fn(self, score, timestep, x_1, condition_mask):
        '''
        Loss function for the score prediction task

        Args:
            score: Predicted score
                    
        The target is the noise added to the data at a specific timestep 
        Meaning the prediction is the approximation of the noise added to the data
        '''
        sigma_t = self.sde.marginal_prob_std(timestep).unsqueeze(1).to(score.device)
        x_1 = x_1.unsqueeze(2).to(score.device)
        condition_mask = condition_mask.unsqueeze(2).to(score.device)
        score = score.unsqueeze(2)

        loss = torch.mean(sigma_t**2 * torch.sum((1-condition_mask)*(x_1+sigma_t*score)**2))

        return loss
    
    #############################################
    # ----- Multi-GPU setup -----
    #############################################
    def _ddp_setup(self, rank, world_size):

        os.environ['MASTER_ADDR'] = 'localhost'
        os.environ["MASTER_PORT"] = "29500"
        
        torch.cuda.set_device(rank)
        dist.init_process_group(
            backend='nccl',
            init_method='env://',
            world_size=world_size,
            rank=rank,
            timeout=datetime.timedelta(seconds=100_000_000)
        )

        self.device = torch.device(f'cuda:{rank}')
        self.SBIm.model.to(self.device)
        self.model = DDP(self.SBIm.model, device_ids=[rank], output_device=rank)

    #############################################
    # ----- Standard Functions -----
    #############################################

    def _get_score(self, x, t, condition_mask):
        """Get score estimate from model"""
        # Get conditional score
        out = self.model(x=x, t=t, c=condition_mask)
        score = self.SBIm.output_scale_function(t, out)
                
        return score

    def _prepare_batch(self, batch, device):
        # Expand data and condition mask to match num_samples
        data, condition_mask, idx = batch
        data = data.to(device)
        condition_mask = condition_mask.to(device)

        return data, condition_mask, idx

    def _prepare_data(self, data, batch_size, rank):
        condition_mask = torch.distributions.bernoulli.Bernoulli(0.33)

        dataset = TensorTupleDataset(data, condition_mask)

        if self.world_size > 1:
            sampler = DistributedSampler(dataset, num_replicas=self.world_size, rank=rank, shuffle=True)
            data_loader = DataLoader(
                dataset, 
                batch_size=batch_size,
                pin_memory=True,
                shuffle=False,
                sampler=sampler
            )
        else:
            data_loader = DataLoader(dataset, batch_size=batch_size, shuffle=True)

        return data_loader

    def _save_checkpoint(self, name="Model_checkpoint"):
        # Save checkpoint
        if self.world_size > 1:
            # Save torch model in case of multi-GPU training
            state_dict = {
                'model_state_dict' : self.model.module.state_dict(),
                'nodes_size': self.SBIm.nodes_size,
                'sde_type': self.SBIm.sde_type,
                'sigma': self.SBIm.sigma,
                'hidden_size': self.SBIm.hidden_size,
                'depth': self.SBIm.depth,
                'num_heads': self.SBIm.num_heads,
                'mlp_ratio': self.SBIm.mlp_ratio
            }
            torch.save(state_dict, f"{self.path}/{name}.pt")
        else:
            self.SBIm.save(path=self.path ,name=name)
